/*
Datei     : radicohw.cpp
Projekt   : XCTL
Subsystem : Detektoren
Autor     : Jan Picard <picard@informatik.hu-berlin.de> 2001-2001
            Institut fuer Informatik,
            Humboldt-Universitaet Berlin
Inhalt    : Implementation der Klasse TRadiconHW
*/

#include  "hwio.h"
#include  "radicohw.h"
#include <mmsystem.h>
#include <memory.h>

static const LNGMSG=25;          // Laenge des Receive/Transmit-Puffers

// Antwortcodes der Controller-Karte
static const ACK=0x06;
static const NAK=0x15;
static const CAN=0x18;           // canceled, impliziert NAK
static const DLE=0x10;           // data line end
static const EOT=0x04;           // end of transmission

// Allgemeine Konstanten
static const TMOUT=3;            // Comm-Timeout
static const long KMGHZ=500000L; // Frequenz-Basis, beeinflusst ExposureTime
static const NCYCL=10;           // Anzahl der Wiederholversuche

static const PROGR=0x6c;         // program ready ???
static const MONIT=0x93;         // monitor controller ???

// nur intern genutzte Funktionscodes
static const SET_PARAMETERS=0x0; // function code
static const UPLOAD_FIRMWARE=0xf;// Programm auf die Controllerkarte laden

static const FIRMWARE_LENGTH = 0x7f0;

/******************************************************************************
--- Konstruktor RadiconHW ---
Initialisiert lediglich die private-Attribute
*******************************************************************************/
TRadiconHW::TRadiconHW( int ctrlPort, int dataPort, BOOL fake )
	 : control( ctrlPort ), data ( dataPort ), fakeDevice( fake )
{
};

/******************************************************************************
--- DWORD CurrentTime ---
Zeit seit dem Start des Systems in Millisekunden ermitteln

Rueckgabewert:  Systemzeit in Millisekunden
*******************************************************************************/
inline DWORD TRadiconHW::CurrentTime()
{
  return timeGetTime();
}

/******************************************************************************
--- int ReadStatus ---
Testen, ob Controller zur Datenuebergabe (senden) bereit

Rueckgabewert:  0  - Controller nicht breit (Datenregister leer)
               !0  - Controller bereit (Datenregister nicht leer)
*******************************************************************************/
inline int TRadiconHW::ReadStatus()
{
  return ( control.readByte() & 0x2 );
}

/******************************************************************************
--- int status_wr ---
Testen ob Controller zur Datenuebernahme (empfangen) bereit

Rueckgabewert: 0  - Controller bereit (Datenregister leer)
              !0  - Controller nicht bereit (Datenregister nicht leer)
*******************************************************************************/
inline int TRadiconHW::WriteStatus()
{
  return ( control.readByte() & 0x1 );
}

/******************************************************************************
--- int ReadData ---
Datenbyte vom Controller empfangen

Parameter:        d - das zu uebertragende Datum

Rueckgabewert:    0 - OK
                 -1 - FEHLER
*******************************************************************************/
int TRadiconHW::ReadData( int &d )
{

  DWORD start = CurrentTime(); // Startzeit sichern
  DWORD now(start);

  int rdy;

  // Busy-Waiting bis Controller empfangsbereit oder Timeout erreicht
  while ((now - start) < TMOUT)
  {
    rdy = ReadStatus();
	 if ( rdy == 0 ) break; // Controller bereit
    now = CurrentTime();   // Systemzeit neu ermitteln

    // Da die Systemzeit ein 32bit-Wert ist, der alle 49,71 Tage
    // ueberlaeuft, wird nach einem Ueberlauf der Startzeitpunkt korrigiert.
    // Im schlimmsten Fall verlaengert sich dadurch
    // der Timeout auf <2*TMOUT
    if ( now < start )
			start = 0 ;
  }

  if (rdy != 0)
		return -1; //Controller nicht bereit -> FEHLER

  d = data.readByte();
  return 0;
}

/******************************************************************************
--- int WriteData ---
Datenbyte zum Controller uebertragen

Parameter:        d - das zu uebertragende Datum

Rueckgabewert:    0 - OK
                 -1 - FEHLER
*******************************************************************************/
int TRadiconHW::WriteData( int d )
{

  DWORD start = CurrentTime(); // Startzeit sichern
  DWORD now(start);

  int rdy;

  // Busy-Waiting bis Controller empfangsbereit oder Timeout erreicht
  while ((now - start) < TMOUT)
  {
    rdy = WriteStatus();
    if ( rdy == 0 ) break; // Controller bereit
    now = CurrentTime();   // Systemzeit neu ermitteln

    // Da die Systemzeit ein 32bit-Wert ist, der alle 49,71 Tage
    // ueberlaeuft, wird nach einem Ueberlauf der Startzeitpunkt korrigiert.
    // Im schlimmsten Fall verlaengert sich dadurch
    // der Timeout auf <2*TMOUT
    if ( now < start )
			start = 0 ;
  }

  if (rdy != 0)
    return -1; //Controller nicht bereit -> FEHLER

  data.writeByte(d);

  return 0;

}

/******************************************************************************
--- int ReceiveMessage ---
Nachricht vom Controller empfangen
(Funktionscode dem Controller uebermitteln und Ergebnisdaten empfangen)

Parameter: int fun             - Funktionscode (Nachricht)
           unsigned char *buf  - Zeiger auf Datenpuffer
           int &len            - Anzahl der empfangenen Daten
           unsigned int maxlen - max. Laenge der zu empfangenden Daten

Rueckgabewert:  0   - OK
               -1   - FEHLER
               -2   - Controller kann die angeforderte Funktion zu diesem
							 Zeitpunkt nicht ausfuehren
                      oder unbekannter Funktionscode
               -3   - Checksummenfehler
               -4   - Daten stehen nicht bereit
*******************************************************************************/
int TRadiconHW::ReceiveMessage(int fun, unsigned char *buf, int &len,
															                    unsigned int maxlen )
{
  if( WriteData(fun) == -1)
    return -1;    // Controller nicht bereit -> FEHLER

  int answer;
  if ( ReadData(answer) == -1)
    return -2;    // Controller antwortet nicht -> FEHLER

  switch(answer)
  {
    case ACK:     // Alles ok
      break;

    case NAK:
      return -2;  // Unbekannter Funktionscode -> FEHLER

    case CAN:     
		return -4;  // Daten stehen nicht bereit

    default:      // unspezifischer Fehler
      return -1;
  }

  int chksum = 0;
  len = 0;

  // Daten byteweise vom Controller empfangen
  // bis maxlen erreicht ist oder das Ende der Uebertragung
  // vom Controller durch DLE signalisiert wird
  while ( len < maxlen )
  {
    if ( ReadData(answer) == -1)
      return -2;      // Controller antwortet nicht -> FEHLER
    chksum += answer; // empfangene Bytes aufsummieren

    if ( answer == DLE )
    {
      // Folgebyte abrufen
      // Folgt EOT, dann ist die Uebertragung beendet.
      // Folgt DLE, Datenwert(10h) uebernehmen (war codiert)
      int answer2;
		if ( ReadData(answer2) == -1)
        return -2;          // Controller antwortet nicht -> FEHLER
      chksum += answer2;    // empfangene Bytes aufsummieren

      if ( answer2 == EOT ) // answer war tatsaechlich DLE
        break;

      if ( answer2 != DLE ) // FEHLER
        return -1;
    }

    *buf = answer;          // empfangenes Byte in den Datenpuffer uebernehmen
	 ++buf; ++len;           // Anzahl der empfangenen Daten inkrementieren
  }

  // letztes Byte vom Controller empfangen
  // (zur Ueberpruefung der Checksumme)
  if ( ReadData(answer) == -1)
    return -2;              // Controller antwortet nicht -> FEHLER

  // Checksumme pruefen
  chksum = ( chksum + answer ) & 0xff;

  // Checksumme != 0 => NAK (wegen Checksummenfehler)
  // an den Controller senden
  if ( chksum != 0 )
  {
    if ( WriteData(NAK) == -1 )
      return -1;
    return -3;              // Funktion beenden
  }

  // Nachricht ACK (Checksumme OK) an Controller uebermitteln
  if ( WriteData(ACK) == -1 )
    return -1;

  return 0;
}

/******************************************************************************
--- int TransmitMessage ---
Nachricht (Funktionscode und Daten) zum Controller uebertragen

Parameter:  int fun            - Funktionscode (Nachricht)
            unsigned char *buf - Zeiger auf Datenpuffer
            unsigned int len   - Anzahl der zu uebertragenden Bytes

Rueckgabewert:  0   - OK
					-1   - FEHLER
               -2   - Controller kann die angeforderte Funktion zu diesem
                      Zeitpunkt nicht ausfuehren
                      oder unbekannter Funktionscode
               -3   - Checksummenfehler
*******************************************************************************/
int TRadiconHW::TransmitMessage(int fun, char const *buf, unsigned int len)
{
  if( WriteData(fun) == -1)
    return -1;      // Controller nicht bereit -> FEHLER

  int answer;
  if ( ReadData(answer) == -1)
    return -2;      // Controller antwortet nicht -> FEHLER

  switch (answer)
  {
    case ACK:       // Alles ok
      break;

    case NAK:
      return -2;

    case PROGR:
		if ( fun == UPLOAD_FIRMWARE )
        return 0;

    default:        // Unspezifischer Fehler
      return -1;
  }

	// Fertig, wenn keine Daten zu uebermitteln sind
	if ( len == 0 ) 
		return 0;

  int chksum=0;
  int datum;

  // Datenpuffer byteweise zum Controller uebertragen
  for( int i=0; i<len; ++i )
  {
    datum = (int) *buf;

    if ( WriteData( datum ) == -1 )
      return -2;
    chksum += datum;

    // Jedes 0x10 doppelt senden
	 if ( datum == DLE )
    {
      if ( WriteData( datum ) == -1 )
        return -2;
      chksum += datum;
    }

    buf++;  // Pufferzeiger auf naechstes Datum
  }

  // Daten vollstaendig uebertragen -> DLE+EOT senden
  if ( WriteData( DLE ) == -1 )
	 return -2;
  chksum += datum;
  if ( WriteData( EOT ) == -1 )
    return -2;
  chksum += datum;

  chksum = ~chksum; // => Zweierkomplement bilden
  chksum++;

  // ermittelte Checksumme uebertragen
  if ( WriteData( chksum ) == -1 )
    return -2;
  chksum += datum;

  if ( ReadData( answer ) == -1 )
    return -2;

  if ( answer == NAK ) // Checksummenfehler
      return -3;

  return 0;
}

/*******************************************************************************
--- SetParameters ---
Geraeteparameter in den Controller schreiben

Rueckgabewerte:     0 = Operation erfolgreich
                   -1 = Kommunikationsfehler
                   -2 = Controller kann die angeforderte Funktion zu diesem
                        Zeitpunkt nicht ausfuehren oder Fehler
******************************************************************************/

int TRadiconHW::SetParameters(
				unsigned short upperThresh, unsigned short lowerThresh,
				int highVoltage, double exposureTime, unsigned long impulseCount,
				BOOL soundEnabled, int energyIntegral, int Interrupt )
{
  if (fakeDevice)
	 return 0;										// Nur simulieren

  char bufmsg[LNGMSG];
  memset( bufmsg, '\0', LNGMSG );

  bufmsg[ 0] = (char) energyIntegral;
  bufmsg[ 1] = (soundEnabled == TRUE) ? '7' : '8';

  bufmsg[ 7] = (char) highVoltage;
  bufmsg[ 8] = (char) (( highVoltage >> 8 ) & 0xff);

  bufmsg[ 9] = (char) lowerThresh;
  bufmsg[10] = (char) (( lowerThresh >> 8 ) & 0xff);

  bufmsg[11] = (char) upperThresh;
  bufmsg[12] = (char) (( upperThresh >> 8 ) & 0xff);

  ulong maxcounts = (ulong) exposureTime * (ulong) KMGHZ;
  bufmsg[13] = (char) ( maxcounts & 0xff );
  bufmsg[14] = (char) ( ( maxcounts >>  8 ) & 0xff );
  bufmsg[15] = (char) ( ( maxcounts >> 16 ) & 0xff );
  bufmsg[16] = (char) ( ( maxcounts >> 24 ) & 0xff );

  bufmsg[17] = (char) ( impulseCount & 0xff );
  bufmsg[18] = (char) ( ( impulseCount >>  8 ) & 0xff );
  bufmsg[19] = (char) ( ( impulseCount >> 16 ) & 0xff );
  bufmsg[20] = (char) ( ( impulseCount >> 24 ) & 0xff );

  bufmsg[21] = (char) ( Interrupt & 0xff );

  int answer;
  for ( int cycles=0; cycles < NCYCL; ++cycles )
  {
	 answer = TransmitMessage ( SET_PARAMETERS, bufmsg, LNGMSG );

	 if ( answer == 0 ) return 0;   // Uebertragung erfolgreich
	 if ( answer == -2 ) return -2; // Controller reagiert nicht
											  // oder Fktn. zur Zeit nicht moeglich
  }

  return -1;    // Timeout erreicht
}

/*******************************************************************************
-- SetOperationMode ---
Die Arbeit des Controllers kontrollieren

Rueckgabe     0 = Operation erfolgreich
				 -1 = Kommunikationsfehler
				 -2 = Controller kann die angeforderte Funktion zu diesem
						Zeitpunkt nicht ausfuehren oder Fehler
*******************************************************************************/
int TRadiconHW::SetOperationMode( operationMode fun )
{
  if (fakeDevice)
	 return 0;									     // Nur simulieren

  int answer;
  for ( int cycles=0; cycles < NCYCL; ++cycles )
  {
	 answer = TransmitMessage ( fun );

	 if ( answer == 0 ) return 0;   // Uebertragung erfolgreich
	 if ( answer == -2 ) return -2; // Controller reagiert nicht
											  // oder Fktn. zur Zeit nicht moeglich
  }

  return -1;                       // Timeout erreicht
}

/*******************************************************************************
--- GetValues ---
Messwerte vom Controller auslesen

Rueckgabe:     0 = Operation erfolgreich
				  -1 = Kommunikationsfehler
				  -2 = Controller kann die angeforderte Funktion zu diesem
						 Zeitpunkt nicht ausfuehren oder Fehler
				  -4 = Daten sind nicht bereit
******************************************************************************/
int TRadiconHW::GetValues( readoutMode mode, double &exposureTime,
																			 unsigned long &impulseCount )
{
  unsigned char bufmsg[LNGMSG];

  if (fakeDevice)										// nur simulieren
  {
	 exposureTime = 0.0;
	 impulseCount = 0;
	 return 0;
  }

  int answer, msglen;
  for ( int cycles=0; cycles < NCYCL; ++cycles )
  {
	answer = ReceiveMessage ( mode, bufmsg, msglen, LNGMSG );

	if ( answer == 0 ) break;        // Uebertragung erfolgreich
	if ( answer == -2 || answer == -4)
	 return answer;
  }

  if ( answer != 0 ) return -1;     // Timeout erreicht

  // Antwort decodieren
  unsigned char *ptr=bufmsg;        // Iterator ueber der Nachricht
  ulong tmp(0);                     // Zwischenspeicher

  // ExposureTime auslesen und per bitshifts einen
  // long-Wert erstellen, aus dem anschliessend das Ergebnis
  // berechnet wird
  tmp += (ulong)(*ptr);
  ptr++;
  tmp += ( ulong(*ptr << 8) );
  ptr++;
  tmp += ( ulong(*ptr << 16) );
  ptr++;
  tmp += ( ulong(*ptr << 24) );
  ptr++;

  exposureTime = ((double) tmp) / ((double) KMGHZ);

  // impulseCount auslesen
  tmp = ulong(*ptr);
  ptr++;
  tmp += ( ulong(*ptr << 8) );
  ptr++;
  tmp += ( ulong(*ptr << 16) );
  ptr++;
  tmp += ( ulong(*ptr << 24) );
  ptr++;

  impulseCount=tmp;

  return 0;
}

/*******************************************************************************
--- FirmwareUpload ---
Programm in den Controller laden

Rueckgabe:     0 = Operation erfolgreich
              -1 = Kommunikationsfehler
              -2 = Controller kann die angeforderte Funktion zu diesem
                   Zeitpunkt nicht ausfuehren oder Fehler
              -3 = Programm zu lang
*******************************************************************************/
int TRadiconHW::FirmwareUpload ( char const *bufmsg, unsigned int len)
{
  // In dieser Methode werden mehrfach aus unklaren Gruenden
  // Bytes vom Controller gelesen. Ich habe diese Verhalten
  // aus den alten Quellen uebernommen, da ich das Kommunikations-
  // protokoll des Radicon-Controllers nicht kenne.

  if (fakeDevice)						// nur simulieren
   return 0;

  if ( len != FIRMWARE_LENGTH )
    return -3;              // Groesse des Firmware-Programm inkorrekt

  // Nach Controller-Reset TMOUT warten
  // Das scheint eine Sicherheitsmassnahmen zu sein,
  // die dem Controller eine gewisse Zeit einraeumt,
  // bevor er korrekte Werte annimmt und liefert.
  DWORD start = CurrentTime();    // Startzeit merken
  DWORD now = start;
  for (;;)
  {
    now = CurrentTime();

    // Da die Systemzeit ein 32bit-Wert ist, der alle 49,71 Tage
    // ueberlaeuft, wird nach einem Ueberlauf der Startzeitpunkt korrigiert.
    // Im schlimmsten Fall verlaengert sich dadurch
    // der Timeout auf <2*TMOUT
    if ( now < start )
      start = now;

    if ( now - start >= TMOUT )
      break;
  }

  // Ein Byte vom Controller lesen - Grund???
  int d;
  ReadData(d);

  // Controller stoppen
  SetOperationMode(stopDevice);

  // Programm zum Controller senden
  int answer;
  for ( int cycles=0; cycles < NCYCL; ++cycles )
  {
    answer = TransmitMessage ( UPLOAD_FIRMWARE, bufmsg, len );

    if ( answer == 0 ) break;       // Uebertragung erfolgreich
    if ( answer == -2 ) return -2;  // Controller reagiert nicht
                                    // oder Fktn. zur Zeit nicht moeglich
  }

  if ( answer != 0 ) return -1;     // Timeout erreicht

  // Der folgende Abschnitt scheint nicht sehr sinnvoll,
  // liest aber im Falle d==MONIT wenigstens ein Byte vom Controller.
  // Ansonsten gilt das zum Beginn der Methode gesagte.
  if ( d == MONIT )
    if ( ReadData(d) == 0 )
      if ( d == PROGR )
        return 0;

  return 0;
}

/*******************************************************************************
--- Reset ---
Controller zuruecksetzen
*******************************************************************************/
inline void TRadiconHW::Reset()
{
  if (! fakeDevice)
    control.writeByte(0);
}

inline BOOL TRadiconHW::FakeDevice()
{
  return fakeDevice;
}

inline void TRadiconHW::FakeDevice(BOOL fake)
{
  fakeDevice=fake;
}

