#ifndef __M_LAYER_H
#define __M_LAYER_H



/*
 * High Level layer der Motoransteuerung.
 * Funktionen von ausserhalb sollten diese benutzen und nicht die
 * aus motors.h
 *
 * mFunctionName  = funktion benutzt den aktivierten Motor
 * mlFunctionName = funktion benutzt den Motor mit der uebergebenen
 *                  MotorID (mid)
 * Achse = ein Motor, der eine bestimmte Achse bewegt
 *
 * Viele hier benutzte Typen werden in comhead.h definiert, auf
 * jeden Fall einmal anschauen.
 * In m_motcom.h stehen einige Kurzroutinen!!
 *
 * Kommentare: Jens-Uwe Rumstich, 18. Februar 1999
 *             Stefan Ltzkendorf, Februar 2001
 */

#ifdef __cplusplus
extern "C" {
#endif
// prototypes for export functions

/*******************************************************************************
 ******************** MotorListen-Funktionen ***********************************
 ******************************************************************************/

/* wurde Motors.DLL geladen und initialisiert?
 * TRUE  = ja,
 * FALSE = noe
 */
BOOL      WINAPI mlIsServerOK( void );

/*******************************************************************************
 ************************ Motor-Funktionen *************************************
 ******************************************************************************/

/* Bewegt aktuellen Motor relativ zur aktuellen Position um eine
 * bestimmte Distanz.
 * Distanz wird im Winkelmass (Nutzereinheiten) uebergeben
 */
BOOL      WINAPI mMoveByDistance( double distance);

BOOL      WINAPI mSetLine( int, BOOL );

/* Setzen der Defaults fuer Winkel (dAngleBias auf 0 ) und der
 * (erlaubten ?)Winkelbereiche (SetAngleRange). Die Werte brauchen nicht
 * uebergeben zu werden.
 */
void      WINAPI mSetAngleDefault( void );

/*******************************************************************************
 ****** "MoveScan"-Funktionen **************************************************
 *******************************************************************************
 *
 * Ein MoveScan setzt den Motor mit den aktuellen Parametern
 * in Bewegung und tastet die Position in regelmigen Abstnden
 * ab. Der Entstehende Vektor von Positionsangaben steht danach zur
 * Verfgung.
 *
 */

/* startet einen MoveScan fr den aktuellen Motor
 */
void      WINAPI mStartMoveScan( int, int );

/*******************************************************************************
 ************************ Motor-Simulation *************************************
 ******************************************************************************/
// bit 0 - hardware
// bit 1 - simulation
enum TSimulationType {
  no_simulation   = 1,
  simulation_only = 2,
  test_simulation = 3
};

/* legt Simulationstyp fest
 *
 * t == no_simulation
 *      keine Simulation, evtl. registierte callbacks werden 
 *      ignoriert, es wird nur die Hardware angesteuert
 * t == simulation_only
 *      keine Ansteuerung der Hardware, es werden fr Schreib- und Lese-
 *      Operationen die registirerten callbacks verwendet
 * t == test_simulation
 *      es wird erst die Hardware angesteuert und dann die ber die 
 *      registierten callbacks die Simulation, Lese-Operationen liefern
 *      zustzlich die Hardwarewerte fr den Vergleich mit der Simulation
 */
void WINAPI msSetSimulationType(TSimulationType t);

/* Typen fr callback-Funktionen zur Kommunikation mit einer
 * Motornensimulation
 * 
 * GET-Funktionen sind Lese-Operationen aus Sicht der motors.dll
 * PUT-Funktionen sind Schreib-Operationen 
 *
 * die Parameter der GET-funktionen sind
 *   - die Quelle (addresse oder port)
 *   - der Hardware-Vergleichs-Wert falls der 
 *     Simulationstype test_simulation ist
 * die GET-Funktionen liefern den von der Simulation 
 * generierten Wert
 *
 * die Parameter der PUT-Funktionen sind
 *   - Ziel der Schreiboperation (addresse oder port)
 *   - der zu schreibende Wert
 */ 
typedef char WINAPI (*T812ISA_GET_CALLBACK) (char* addr, char hw_value);
typedef void WINAPI (*T812ISA_PUT_CALLBACK) (char* addr, char put);
typedef int  WINAPI (*T832_GET_CALLBACK) (unsigned port, int hw_value);
typedef void WINAPI (*T832_PUT_CALLBACK) (unsigned port, int put);

/* Funktionen zum Registrieren oben beschriebener callback-Funktionen
 */
void WINAPI msRegister_C812ISA_Get(T812ISA_GET_CALLBACK);
void WINAPI msRegister_C812ISA_Put(T812ISA_PUT_CALLBACK);
void WINAPI msRegister_C832_Get(T832_GET_CALLBACK);
void WINAPI msRegister_C832_Put(T832_PUT_CALLBACK);


#ifdef __cplusplus
 }
#endif

#ifdef __WIN32__
/* verschiedene Type-Definitionen (Function-Casts)
 */
typedef BOOL   WINAPI   (*TInitializeMotorsDLL)( void );
typedef BOOL   WINAPI   (*TSetAxis)( int );
typedef int    WINAPI   (*TGetAxis)( void );
typedef BOOL   WINAPI   (*TSetLine)( int, BOOL );
typedef BOOL   WINAPI   (*TIsMoveFinish)( void );
typedef BOOL   WINAPI   (*TGetDistance)( double& );
typedef BOOL   WINAPI   (*TIsServerOK)( void );
typedef BOOL   WINAPI   (*TMoveToDistance)( double );
typedef BOOL   WINAPI   (*TMoveByDistance)( double );
typedef BOOL   WINAPI   (*TSetAxisWithName)( TAxisType );
typedef int    WINAPI   (*TGetScanSize)( void );
typedef LPLONG WINAPI   (*TGetMoveScan)( void );
typedef void   WINAPI   (*TStartMoveScan)( int, int );
typedef int    WINAPI   (*TGetMoveFinishIdx)( void );
typedef void   CALLBACK (*TSavePosition)( UINT, UINT, DWORD, DWORD, DWORD );

/* table with several functions
 */
typedef struct
{
 public:
  TInitializeMotorsDLL InitializeMotorsDLL;
  TSetAxis             SetAxis;
  TSetAxisWithName     SetAxisWithName;
  TGetAxis             GetAxis;
  TGetScanSize         GetScanSize;
  TGetMoveScan         GetMoveScan;
  TStartMoveScan       StartMoveScan;
  TMoveToDistance      MoveToDistance;
  TMoveByDistance      MoveByDistance;
  TSetLine             SetLine;
  TIsMoveFinish        IsMoveFinish;
  TGetDistance         GetDistance;
  TIsServerOK          IsServerOK;
  TGetMoveFinishIdx    GetMoveFinishIdx;
  TSavePosition        SavePosition;
} TMFunctionTable;
#endif
#endif
// __LastLine__

